@@ -135,6 +135,7 @@ class Csr : public EnableLinOp<Csr<ValueType, IndexType>>,
     friend class Fbcsr<ValueType, IndexType>;
     friend class CsrBuilder<ValueType, IndexType>;
     friend class Csr<to_complex<ValueType>, IndexType>;
+    GKO_ASSERT_SUPPORTED_VALUE_AND_INDEX_TYPE;
 
 public:
     using EnableLinOp<Csr>::convert_to;
@@ -768,6 +769,251 @@ public:
 
     std::unique_ptr<LinOp> conj_transpose() const override;
 
+    /**
+     * Class describing the internal lookup structures created by
+     * multiply_reuse(const Csr*) to recompute a sparse matrix-matrix product
+     * with updated values.
+     */
+    class multiply_reuse_info {
+        friend class Csr;
+
+    public:
+        explicit multiply_reuse_info();
+
+        ~multiply_reuse_info();
+
+        multiply_reuse_info(const multiply_reuse_info&) = delete;
+
+        multiply_reuse_info(multiply_reuse_info&&) noexcept;
+
+        multiply_reuse_info& operator=(const multiply_reuse_info&) = delete;
+
+        multiply_reuse_info& operator=(multiply_reuse_info&&) noexcept;
+
+        /**
+         * Recomputes the sparse matrix-matrix product `out = mtx1 * mtx2` when
+         * only the values of mtx1 and mtx2 changed, but the sparsity patterns
+         * of mtx1, mtx2 and out are unchanged.
+         */
+        void update_values(ptr_param<const Csr> mtx1, ptr_param<const Csr> mtx2,
+                           ptr_param<Csr> out) const;
+
+    private:
+        struct lookup_data;
+
+        explicit multiply_reuse_info(std::unique_ptr<lookup_data> data);
+
+        std::unique_ptr<lookup_data> internal;
+    };
+
+    /**
+     * Computes the sparse matrix product `this * other` on the executor of this
+     * matrix.
+     *
+     * @param other  the matrix with which the product will be computed.
+     *               It needs to be sorted by column indices when using
+     *               OmpExecutor or DpcppExecutor for `this`.
+     * @return  the product of the two matrices, stored on the same executor as
+     *          this matrix.
+     */
+    std::unique_ptr<Csr> multiply(ptr_param<const Csr> other) const;
+
+    /**
+     * Computes the sparse matrix product `this * other` on the executor of this
+     * matrix, and necessary data for value updates:
+     * ```
+     * auto [C, reuse] = A->multiply_reuse(B);
+     * change_values(A, B);
+     * reuse->update_values(A, B, C);
+     * ```
+     *
+     * @param other  the matrix with which the product will be computed.
+     *               It needs to be sorted by column indices when using
+     *               OmpExecutor or DpcppExecutor for `this`.
+     * @return  std::pair containing the product of the two matrices, stored on
+     *          the same executor as this matrix, and a multiply_reuse_info
+     *          object allowing value updates to the output matrix.
+     */
+    std::pair<std::unique_ptr<Csr>, multiply_reuse_info> multiply_reuse(
+        ptr_param<const Csr> other) const;
+
+    /**
+     * Class describing the internal lookup structures created by
+     * multiply_add_reuse to recompute a sparse matrix-matrix product
+     * with updated values.
+     */
+    class multiply_add_reuse_info {
+        friend class Csr;
+
+    public:
+        explicit multiply_add_reuse_info();
+
+        ~multiply_add_reuse_info();
+
+        multiply_add_reuse_info(const multiply_add_reuse_info&) = delete;
+
+        multiply_add_reuse_info(multiply_add_reuse_info&&) noexcept;
+
+        multiply_add_reuse_info& operator=(const multiply_add_reuse_info&) =
+            delete;
+
+        multiply_add_reuse_info& operator=(multiply_add_reuse_info&&) noexcept;
+
+        /**
+         * Recomputes the sparse matrix-matrix product
+         * `out = scale_mult * mtx * mtx_mult + scale_add * mtx_add` when only
+         * the values of mtx, scale_mult, mtx_mult, scale_add, mtx_add changed,
+         * but the sparsity patterns of mtx, mtx_mult, mtx_add and out are
+         * unchanged.
+         */
+        void update_values(ptr_param<const Csr> mtx,
+                           ptr_param<const Dense<value_type>> scale_mult,
+                           ptr_param<const Csr> mtx_mult,
+                           ptr_param<const Dense<value_type>> scale_add,
+                           ptr_param<const Csr> mtx_add,
+                           ptr_param<Csr> out) const;
+
+    private:
+        struct lookup_data;
+
+        explicit multiply_add_reuse_info(std::unique_ptr<lookup_data> data);
+
+        std::unique_ptr<lookup_data> internal;
+    };
+
+    /**
+     * Computes the sparse matrix product
+     * `scale_mult * this * mtx_mult + scale_add * mtx_add` on the executor of
+     * this matrix.
+     *
+     * @param scale_mult  the scalar by which the matrix product will be scaled.
+     * @param mtx_mult    the matrix with which the product will be computed. It
+     *                    needs to be sorted by column indices when using
+     *                    OmpExecutor or DpcppExecutor for `this`.
+     * @param scale_add   the scalar by which the matrix mtx_add will be scaled.
+     * @param mtx_add     the matrix which will be added to the product, scaled
+     *                    by scale_add.
+     * @return  the result of the computation, stored on the same executor as
+     *          this matrix.
+     */
+    std::unique_ptr<Csr> multiply_add(
+        ptr_param<const Dense<value_type>> scale_mult,
+        ptr_param<const Csr> mtx_mult,
+        ptr_param<const Dense<value_type>> scale_add,
+        ptr_param<const Csr> mtx_add) const;
+
+    /**
+     * Computes the sparse matrix product
+     * `scale_mult * this * mtx_mult + scale_add * mtx_add` on the executor of
+     * this matrix, and necessary data for value updates:
+     * ```
+     * auto [result, reuse] = mtx->multiply_add_reuse(sm, mm, sa, ma);
+     * change_values(mtx, sm, mm, sa, ma);
+     * reuse->update_values(mtx, sm, mm, sa, ma, result);
+     * ```
+     *
+     * @param scale_mult  the scalar by which the matrix product will be scaled.
+     * @param mtx_mult    the matrix with which the product will be computed. It
+     *                    needs to be sorted by column indices when using
+     *                    OmpExecutor or DpcppExecutor for `this`.
+     * @param scale_add   the scalar by which the matrix mtx_add will be scaled.
+     * @param mtx_add     the matrix which will be added to the product, scaled
+     *                    by scale_add.
+     * @return  std::pair containing the result of the computation, stored on
+     *          the same executor as this matrix, and a multiply_add_reuse_info
+     *          object allowing value updates to the output matrix.
+     */
+    std::pair<std::unique_ptr<Csr>, multiply_add_reuse_info> multiply_add_reuse(
+        ptr_param<const Dense<value_type>> scale_mult,
+        ptr_param<const Csr> mtx_mult,
+        ptr_param<const Dense<value_type>> scale_add,
+        ptr_param<const Csr> mtx_add) const;
+
+    /**
+     * Class describing the internal lookup structures created by
+     * scale_add_reuse to recompute a sparse matrix-matrix sum
+     * with updated values.
+     */
+    class scale_add_reuse_info {
+        friend class Csr;
+
+    public:
+        explicit scale_add_reuse_info();
+
+        ~scale_add_reuse_info();
+
+        scale_add_reuse_info(const scale_add_reuse_info&) = delete;
+
+        scale_add_reuse_info(scale_add_reuse_info&&) noexcept;
+
+        scale_add_reuse_info& operator=(const scale_add_reuse_info&) = delete;
+
+        scale_add_reuse_info& operator=(scale_add_reuse_info&&) noexcept;
+
+        /**
+         * Recomputes the sparse matrix-matrix sum
+         * `out = scale1 * mtx1 + scale2 * mtx2` when only the values of
+         * mtx1, scale1, mtx2, scale2 changed, but the sparsity patterns of
+         * mtx1, mtx2 and out are unchanged.
+         */
+        void update_values(ptr_param<const Dense<value_type>> scale1,
+                           ptr_param<const Csr> mtx1,
+                           ptr_param<const Dense<value_type>> scale2,
+                           ptr_param<const Csr> mtx2, ptr_param<Csr> out) const;
+
+    private:
+        struct lookup_data;
+
+        explicit scale_add_reuse_info(std::unique_ptr<lookup_data> data);
+
+        std::unique_ptr<lookup_data> internal;
+    };
+
+    /**
+     * Computes the sparse matrix sum
+     * `scale_this * this + scale_other * mtx_add` on the executor of this
+     * matrix. This matrix needs to be sorted by column index, otherwise the
+     * result will be incorrect.
+     *
+     * @param scale_this   the scalar by which this matrix will be scaled.
+     * @param scale_other  the scalar by which this matrix will be scaled.
+     * @param mtx_other    the matrix which will be added to this, scaled by
+     *                     scale_other. It needs to be sorted by column index,
+     *                     otherwise the result will be incorrect.
+     * @return  the result of the computation, stored on the same executor as
+     *          this matrix.
+     */
+    std::unique_ptr<Csr> scale_add(
+        ptr_param<const Dense<value_type>> scale_this,
+        ptr_param<const Dense<value_type>> scale_other,
+        ptr_param<const Csr> mtx_other) const;
+
+    /**
+     * Computes the sparse matrix sum
+     * `scale_this * this + scale_other * mtx_add` on the executor of this
+     * matrix, and necessary data for value updates:
+     * ```
+     * auto [result, reuse] = mtx->add_scale_reuse(alpha, beta, mtx2);
+     * change_values(alpha, mtx, beta, mtx2);
+     * reuse->update_values(alpha, mtx, beta, mtx2, result);
+     * ```
+     * This matrix needs to be sorted by column index, otherwise the
+     * result will be incorrect.
+     *
+     * @param scale_this   the scalar by which this matrix will be scaled.
+     * @param scale_other  the scalar by which this matrix will be scaled.
+     * @param mtx_other    the matrix which will be added to this, scaled by
+     *                     scale_other. It needs to be sorted by column index,
+     *                     otherwise the result will be incorrect.
+     * @return  std::pair containing the result of the computation, stored on
+     *          the same executor as this matrix, and a scale_add_reuse_info
+     *          object allowing value updates to the output matrix.
+     */
+    std::pair<std::unique_ptr<Csr>, scale_add_reuse_info> add_scale_reuse(
+        ptr_param<const Dense<value_type>> scale_this,
+        ptr_param<const Dense<value_type>> scale_other,
+        ptr_param<const Csr> mtx_other) const;
+
     /**
      * A struct describing a transformation of the matrix that reorders the
      * values of the matrix into the transformed matrix.
@@ -777,7 +1023,7 @@ public:
         explicit permuting_reuse_info();
 
         /** Creates a reuse info structure from its value permutation. */
-        permuting_reuse_info(
+        explicit permuting_reuse_info(
             std::unique_ptr<Permutation<index_type>> value_permutation);
 
         /**
@@ -801,8 +1047,9 @@ public:
      * change_values(matrix);
      * reuse->update_values(matrix, transposed);
      * ```
-     * @return the reuse info struct that can be used to update values in the
-     *         transposed matrix.
+     * @return an std::pair consisting of the transposed matrix and a reuse info
+     *         struct that can be used to update values in the transposed
+     *         matrix.
      */
     std::pair<std::unique_ptr<Csr>, permuting_reuse_info> transpose_reuse()
         const;
